import Foundation

extension String: JSONDecodable, JSONEncodable {
  public init(jsonValue value: JSONValue) throws {
    switch value {
    case let string as String:
        self = string
    case let int as Int:
        self = String(int)
    default:
        throw JSONDecodingError.couldNotConvert(value: value, to: String.self)
    }
  }

  public var jsonValue: JSONValue {
    return self
  }
}

extension Int: JSONDecodable, JSONEncodable {
  public init(jsonValue value: JSONValue) throws {
    guard let number = value as? NSNumber else {
      throw JSONDecodingError.couldNotConvert(value: value, to: Int.self)
    }
    self = number.intValue
  }

  public var jsonValue: JSONValue {
    return self
  }
}

extension Float: JSONDecodable, JSONEncodable {
  public init(jsonValue value: JSONValue) throws {
    guard let number = value as? NSNumber else {
      throw JSONDecodingError.couldNotConvert(value: value, to: Float.self)
    }
    self = number.floatValue
  }

  public var jsonValue: JSONValue {
    return self
  }
}

extension Double: JSONDecodable, JSONEncodable {
  public init(jsonValue value: JSONValue) throws {
    guard let number = value as? NSNumber else {
      throw JSONDecodingError.couldNotConvert(value: value, to: Double.self)
    }
    self = number.doubleValue
  }

  public var jsonValue: JSONValue {
    return self
  }
}

extension Bool: JSONDecodable, JSONEncodable {
  public init(jsonValue value: JSONValue) throws {
    guard let bool = value as? Bool else {
        throw JSONDecodingError.couldNotConvert(value: value, to: Bool.self)
    }
    self = bool
  }

  public var jsonValue: JSONValue {
    return self
  }
}

extension RawRepresentable where RawValue: JSONDecodable {
  public init(jsonValue value: JSONValue) throws {
    let rawValue = try RawValue(jsonValue: value)
    if let tempSelf = Self(rawValue: rawValue) {
      self = tempSelf
    } else {
      throw JSONDecodingError.couldNotConvert(value: value, to: Self.self)
    }
  }
}

extension RawRepresentable where RawValue: JSONEncodable {
  public var jsonValue: JSONValue {
    return rawValue.jsonValue
  }
}

extension Optional where Wrapped: JSONDecodable {
  public init(jsonValue value: JSONValue) throws {
    if value is NSNull {
      self = .none
    } else {
      self = .some(try Wrapped(jsonValue: value))
    }
  }
}

// Once [conditional conformances](https://github.com/apple/swift-evolution/blob/master/proposals/0143-conditional-conformances.md) have been implemented, we should be able to replace these runtime type checks with proper static typing

extension Optional: JSONEncodable {
  public var jsonValue: JSONValue {
    switch self {
    case .none:
      return NSNull()
    case .some(let wrapped as JSONEncodable):
      return wrapped.jsonValue
      
    // WORKAROUND: For reasons I don't totally understand, when the underlying type is `Any`,
    // even though all of these conform to `JSONEncodable`, the `as JSONEncodable` above
    //  fails, and we need to handle them individually.
    case .some(let wrapped as String):
      return wrapped.jsonValue
    case .some(let wrapped as Int):
      return wrapped.jsonValue
    case .some(let wrapped as Double):
      return wrapped.jsonValue
    case .some(let wrapped as Bool):
      return wrapped.jsonValue
    case .some(let wrapped as [String: Any?]):
      return wrapped.jsonValue
    case .some(let wrapped as [Any?]):
      return wrapped.jsonValue
    default:
      fatalError("Optional is only JSONEncodable if Wrapped is")
    }
  }
}

extension Dictionary: JSONEncodable {
  public var jsonValue: JSONValue {
    return jsonObject
  }

  public var jsonObject: JSONObject {
    var jsonObject = JSONObject(minimumCapacity: count)
    for (key, value) in self {
      if case let (key as String, value as JSONEncodable) = (key, value) {
        jsonObject[key] = value.jsonValue
      } else {
        fatalError("Dictionary is only JSONEncodable if Value is (and if Key is String)")
      }
    }
    return jsonObject
  }
}

extension Dictionary: JSONDecodable {
    public init(jsonValue value: JSONValue) throws {
        guard let dictionary = value as? Dictionary else {
            throw JSONDecodingError.couldNotConvert(value: value, to: Dictionary.self)
        }
        
        self = dictionary
    }
}

extension Array: JSONEncodable {
  public var jsonValue: JSONValue {
    return map() { element -> (JSONValue) in
      if case let element as JSONEncodable = element {
        return element.jsonValue
      } else {
        fatalError("Array is only JSONEncodable if Element is")
      }
    }
  }
}

// Example custom scalar

extension URL: JSONDecodable, JSONEncodable {
  public init(jsonValue value: JSONValue) throws {
    guard let string = value as? String else {
      throw JSONDecodingError.couldNotConvert(value: value, to: URL.self)
    }

    if let url = URL(string: string) {
        self = url
    } else {
        throw JSONDecodingError.couldNotConvert(value: value, to: URL.self)
    }
  }

  public var jsonValue: JSONValue {
    return self.absoluteString
  }
}
